#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <math.h>

/* fsb_mpeg 0.1 */
/* parse some number of consecutive padded MPEG frames out of a file,
   deinterleave them into several streams */

/* MPEG header info from http://mpgedit.org/mpgedit/mpeg_format/mpeghdr.htm */

struct mpeg_header
{
    unsigned version        : 2;
    unsigned layer          : 2;
    unsigned crc_missing    : 1;
    unsigned bitrate_idx    : 4;
    unsigned samplerate_idx : 2;
    unsigned padding        : 1;
    unsigned private        : 1;
    unsigned channel_mode   : 2;
    unsigned js_mode_ext    : 2;
    unsigned copyrighted    : 1;
    unsigned original       : 1;
    unsigned emphasis       : 2;
};

/* return 0 on success, -1 on failure */
int load_header(struct mpeg_header *h, const uint8_t *h_in)
{
    memset(h, 0, sizeof(*h));

    /* 31-21: check for sync */
    if (h_in[0] != 0xff || (h_in[1] & 0xe0) != 0xe0)
    {
        return -1;
    }

    /* 20,19: version */
    h->version = (h_in[1] & 0x18) >> 3;

    /* 18,17: layer */
    h->layer = (h_in[1] & 0x06) >> 1;

    /* 16: protected by crc? */
    h->crc_missing = (h_in[1] & 0x01);

    /* 15-12: bitrate index */
    h->bitrate_idx = (h_in[2] & 0xf0) >> 4;

    /* 11,10: sampling rate frequency index */
    h->samplerate_idx = (h_in[2] & 0x0c) >> 2;

    /* 9: padding? */
    h->padding = (h_in[2] & 0x02) >> 1;

    /* 8: private bit */
    h->private = (h_in[2] & 0x01);

    /* 7,6: channel mode */
    h->channel_mode = (h_in[3] & 0xc0) >> 6;

    /* 5,4: mode extension for joint stereo */
    h->js_mode_ext = (h_in[3] & 0x30) >> 4;

    /* 3: copyrighted */
    h->copyrighted = (h_in[3] & 0x08) >> 3;

    /* 2: original */
    h->original = (h_in[3] & 0x04) >> 2;

    /* 1,0: emphasis */
    h->emphasis = (h_in[3] & 0x03);

    return 0;
}

struct mpeg_frame_info
{
    unsigned version;
    unsigned layer;
    unsigned bitrate;
    unsigned samplerate;
    unsigned frame_size;    /* samples */
    unsigned frame_length;  /* bytes */
    unsigned channels;
    char crc_present, private_bit, copyrighted, original;
};

/* return 0 for valid header, -1 for invalid */
int decode_header(struct mpeg_frame_info *f, const struct mpeg_header *h)
{
    memset(f, 0, sizeof(*f));

    if (1 == h->version || 0 == h->layer)
    {
        /* reserved */
        return -1;
    }

    f->version = h->version;
    f->layer = h->layer;

    int bitrate;

    /* look up bitrate */
    switch (f->version)
    {
        case 0: /* Version 2.5 */
        case 2: /* Version 2 */
        {
            /* not dealing with V2 now */
            return -1;
        }
        case 3: /* Version 1 */
        {
            static const int v1_bitrates[4][16] =
            { {},   /* reserved */
              { 0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, -1 }, /* Layer III */
              { 0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, -1 }, /* Layer II */
              { 0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1 }  /* Layer I */ };

            bitrate = v1_bitrates[h->layer][h->bitrate_idx];
            break;
        }
        default:
        {
            return -1;
        }
    }

    if (-1 == bitrate || 0 == bitrate)
    {
        /* reject invalid or freeform */
        return -1;
    }

    f->bitrate = bitrate;

    int samplerate;
    /* look up sampling rate */
    switch (h->version)
    {
        case 0: /* Version 2.5 */
        case 1: /* Version 2 */
        {
            return -1;
        }
        case 3: /* Version 1 */
        {
            static const int v1_samplerates[4] = {44100, 48000, 32000, -1};

            samplerate = v1_samplerates[h->samplerate_idx];
            break;
        }
        default:
        {
            return -1;
        }
    }

    if (-1 == samplerate)
    {
        /* reject reserved */
        return -1;
    }

    f->samplerate = samplerate;

    /* look up frame size */
    static const int frame_sizes[4] = { -1, 1152, 1152, 384 };
    f->frame_size = frame_sizes[h->layer];

    /* calculate frame length */
    int frame_length;
    switch (h->layer)
    {
        case 3: /* Layer I */
            frame_length = (12l * bitrate * 1000l / samplerate + h->padding) * 4;
            break;
        case 1: /* Layer III */
        case 2: /* Layer II */
            frame_length = 144l * bitrate * 1000l / samplerate + h->padding;
            break;
        default:
            return -1;
    }

    //frame_length += 4;      /* header (included) */
    //if (!h->crc_missing)
    //{
    //    frame_length += 2;  /* CRC (16 bits following header) (included?)*/
    //}
    f->frame_length = frame_length;

    switch (h->channel_mode)
    {
        case 0: /* Stereo */
        case 1: /* Joint stereo */
        case 2: /* Dual channel */
            f->channels = 2;
            break;
        case 3: /* Single channel */
            f->channels = 1;
            break;
    }

    f->crc_present = !h->crc_missing;
    f->copyrighted = h->copyrighted;
    f->private_bit = h->private;
    f->original = h->original;

    return 0;
}

void describe_header(const struct mpeg_frame_info *f, FILE *outfile)
{
    static const char *version_strs[4] = {"2.5", "reserved", "2", "1"};
    static const char *layer_strs[4] = {NULL, "III", "II", "I"};

    fprintf(outfile, "MPEG Version %s Layer %s %u kbps %u Hz %s\n",
            version_strs[f->version], layer_strs[f->layer], f->bitrate,
            f->samplerate, ( 2 == f->channels ? "stereo" : "mono" ) );

    fprintf(outfile, "Frame is %u samples, %u bytes\n", f->frame_size,
            f->frame_length);
    fprintf(outfile, "Flags:");

    const char *sep = " ";
    if (f->crc_present)
    {
        fprintf(outfile, "%sCRC present", sep);
        sep = ", ";
    }
    if (f->private_bit)
    {
        fprintf(outfile, "%sPrivate Bit Set", sep);
        sep = ", ";
    }
    if (f->copyrighted)
    {
        fprintf(outfile, "%sCopyrighted", sep);
        sep = ", ";
    }
    if (f->original)
    {
        fprintf(outfile, "%sOriginal", sep);
        sep = ", ";
    }

    printf("\n");
}


int main(int argc, char **argv)
{
    struct mpeg_header header;
    struct mpeg_frame_info info;

    uint8_t buf[10000]; /* largest MPEG audio frame is around 8k */

    FILE *infile;
    FILE **outfiles;
    char **outfile_names;
    long stream_count;
    long offset;
    long *sample_totals;

    if (4 != argc && 3 != argc)
    {
        fprintf(stderr, "fsb_mpeg 0.1 by hcs\n");
        fprintf(stderr, "usage: fsb_mpeg infile.fsb outfile_prefix [stream_count]\n");
        return 1;
    }

    infile = fopen(argv[1], "rb");

    if (!infile)
    {
        perror("fopen input");
        return 1;
    }

    if (4 == argc)
    {
        char *lastc;
        stream_count = strtol(argv[3], &lastc, 10);
        if (
            ((LONG_MIN == stream_count || LONG_MAX == stream_count) &&
            ERANGE == errno) ||
            '\0' == argv[3][0] || 
            '\0' != lastc[0] ||
            0 >= stream_count ||
            UINT_MAX-1 <= stream_count)
        {
            fprintf(stderr, "invalid stream count\n");
            return 1;
        }
    }
    else
    {
        stream_count = 1;
    }

    outfiles = malloc(sizeof(FILE*) * stream_count);
    if (!outfiles)
    {
        perror("malloc");
        return 1;
    }

    outfile_names = malloc(sizeof(char*) * stream_count);
    if (!outfile_names)
    {
        perror("malloc");
        return 1;
    }

    /* name and open output files */
    {
        size_t numberlen = ceil(log10(stream_count+1));
        /* "name_number.mp3\0" */
        size_t namelen = strlen(argv[2]) + 1 + numberlen + 1 + 3 + 1;
        for (long i = 0; i < stream_count; i++)
        {
            outfile_names[i] = malloc(namelen);
            if (!outfile_names[i])
            {
                perror("malloc");
                return 1;
            }

            snprintf(outfile_names[i], namelen, "%s_%0*u.mp3", argv[2],
                    (int)numberlen, (unsigned int)(i+1));
            outfiles[i] = fopen(outfile_names[i], "wb");
            if (!outfiles[i])
            {
                perror("fopen output");
                return 1;
            }
        }
    }

    sample_totals = malloc(sizeof(long) * stream_count);
    if (!sample_totals)
    {
        perror("malloc");
        return 1;
    }
    for (long i = 0; i < stream_count; i++)
    {
        sample_totals[i] = 0;
    }


    /* get file size */
    fseek(infile, 0, SEEK_END);
    long file_size = ftell(infile);

    /* search for a valid MPEG frame */
    for (offset = 0; offset < file_size - 3; offset ++)
    {
        fseek(infile, offset, SEEK_SET);
        if (4 != fread(buf, 1, 4, infile))
        {
            fprintf(stderr, "bad read looking for sync at 0x%lx\n", (unsigned long)offset);
            return 1;
        }

        if (-1 == load_header(&header, buf))
        {
            continue;
        }
        if (-1 == decode_header(&info, &header))
        {
            continue;
        }

        fprintf(stderr, "got sync at 0x%lx\n\n", (unsigned long)offset);
        describe_header(&info, stderr);
        fprintf(stderr, "\n");
        break;
    }

    if (offset == file_size - 3)
    {
        fprintf(stderr, "didn't find a valid MPEG frame\n");
        return 1;
    }

    /* follow the stream */
    long cur_stream = 0;
    while (offset < file_size)
    {
        fseek(infile, offset, SEEK_SET);
        if (4 != fread(buf, 1, 4, infile))
        {
            fprintf(stderr, "bad read of header at 0x%lx\n", (unsigned long)offset);
            return 1;
        }
        if (-1 == load_header(&header, buf))
        {
            fprintf(stderr, "lost sync at 0x%lx (file ends at 0x%lx)\n",
                    (unsigned long)offset, (unsigned long)file_size);
            break;
        }
        if (-1 == decode_header(&info, &header))
        {
            fprintf(stderr, "bad MPEG header at 0x%lx\n", (unsigned long)offset);
            return 1;
        }
        int rounded_length = (info.frame_length + 7) / 8 * 8; /* maybe 16? */
        if (rounded_length > sizeof(buf))
        {
            fprintf(stderr, "I goofed calculating the buffer size\n");
            return 1;
        }
        if (rounded_length-4 != fread(&buf[4], 1, rounded_length-4, infile))
        {
            fprintf(stderr, "bad read at 0x%lx\n", (unsigned long)offset);
            return 1;
        }

        if (info.frame_length !=
                fwrite(buf, 1, info.frame_length, outfiles[cur_stream]))
        {
            perror("fwrite");
            return 1;
        }

        sample_totals[cur_stream] += info.frame_size;

        cur_stream = (cur_stream + 1) % stream_count;
        offset += rounded_length;
    }

    for (long i = 0; i < stream_count; i++)
    {
        fprintf(stderr, "%s: %ld samples\n", outfile_names[i], sample_totals[i]);
        if (EOF == fclose(outfiles[i]))
        {
            perror("fclose");
            return 1;
        }
    }

    fprintf(stderr, "Done!\n");
}
